from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import QAction


class CorrectionAction(QAction):
    """Custom QAction for correcting spelling."""

    actionTriggered = pyqtSignal(str)

    def __init__(self, *args):
        super().__init__(*args)

        self.triggered.connect(self.emitTriggered)

    def emitTriggered(self):
        self.actionTriggered.emit(self.text())



from __future__ import annotations

import re
from typing import TYPE_CHECKING

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QSyntaxHighlighter, QTextCharFormat

if TYPE_CHECKING:
    from pyqt_spellcheck.spellcheckwrapper import SpellCheckWrapper


class SpellCheckHighlighter(QSyntaxHighlighter):
    """Highlighter for spell checking in Qt."""

    wordRegEx = re.compile(r"\b([A-Za-z]{2,})\b")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.speller: SpellCheckWrapper | None = None

        self.misspelledFormat = QTextCharFormat()
        self.misspelledFormat.setUnderlineStyle(QTextCharFormat.SpellCheckUnderline)
        self.misspelledFormat.setUnderlineColor(Qt.GlobalColor.red)

    def highlightBlock(self, text: str | None) -> None:
        if self.speller is None:
            return
        if text is None:
            return

        for word_object in self.wordRegEx.finditer(text):
            if not self.speller.check(word_object.group()):
                self.setFormat(
                    word_object.start(),
                    word_object.end() - word_object.start(),
                    self.misspelledFormat,
                )

    def setSpeller(self, speller: SpellCheckWrapper):
        self.speller = speller



from __future__ import annotations

from PyQt5.QtCore import QEvent, Qt, pyqtSlot
from PyQt5.QtGui import QContextMenuEvent, QMouseEvent
from PyQt5.QtWidgets import QLineEdit, QMenu

from pyqt_spellcheck import util
from pyqt_spellcheck.correction_action import CorrectionAction
from pyqt_spellcheck.spellcheckwrapper import SpellCheckWrapper


class SpellLineEdit(QLineEdit):
    """QLineEdit widget with spell checking."""

    def __init__(self, *args):
        if args and isinstance(args[0], SpellCheckWrapper):
            super().__init__(*args[1:])
            self.speller = args[0]
        else:
            super().__init__(*args)

    def setSpeller(self, speller: SpellCheckWrapper):
        self.speller = speller

    def mousePressEvent(self, event: QMouseEvent | None) -> None:
        if event is None:
            return

        if event.button() == Qt.MouseButton.RightButton:
            event = QMouseEvent(
                QEvent.Type.MouseButtonPress,
                event.pos(),
                Qt.MouseButton.LeftButton,
                Qt.MouseButton.LeftButton,
                Qt.KeyboardModifier.NoModifier,
            )
        super().mousePressEvent(event)

    def contextMenuEvent(self, event: QContextMenuEvent | None) -> None:
        if event is None:
            return

        self.contextMenu = self.createStandardContextMenu()
        if self.contextMenu is None:
            return

        # Select word under cursor
        currentPosition = self.cursorPosition()
        currentText = self.text()
        bounds = util.find_word_bounds(currentText, currentPosition)
        wordToCheck = currentText[bounds[0] : bounds[1]]
        self.setSelection(bounds[0], len(wordToCheck))

        if wordToCheck != "":
            suggestions = self.speller.suggestions(wordToCheck)

            if len(suggestions) > 0:
                self.contextMenu.addSeparator()
                self.contextMenu.addMenu(self.createSuggestionsMenu(suggestions))

            if not self.speller.check(wordToCheck):
                addToDictionary_action = CorrectionAction(
                    "Add to dictionary",
                    self.contextMenu,
                )
                addToDictionary_action.triggered.connect(self.addToDictionary)
                self.contextMenu.addAction(addToDictionary_action)

        self.contextMenu.exec_(event.globalPos())

    def createSuggestionsMenu(self, suggestions: list[str]) -> QMenu:
        suggestionsMenu = QMenu("Change to", self)
        for word in suggestions:
            action = CorrectionAction(word, self.contextMenu)
            action.actionTriggered.connect(self.correctWord)
            suggestionsMenu.addAction(action)

        return suggestionsMenu

    @pyqtSlot(str)
    def correctWord(self, word: str) -> None:
        """Replace the currently selected word with the given word."""
        bounds = (self.selectionStart(), self.selectionEnd())
        self.setSelection(bounds[0], bounds[1] - bounds[0])
        self.insert(word)
        self.setSelection(bounds[0], len(word))

    @pyqtSlot()
    def addToDictionary(self):
        currentPosition = self.cursorPosition()
        currentText = self.text()
        bounds = util.find_word_bounds(currentText, currentPosition)
        newWord = currentText[bounds[0] : bounds[1]]
        self.speller.add(newWord)



from collections.abc import Callable

from enchant import DictWithPWL
from PyQt5.QtCore import QTemporaryFile


class SpellCheckWrapper:
    """Wrapper for enchant spell checking library."""

    def __init__(
        self,
        personal_word_list: list[str],
        addToDictionary: Callable[[str], None],
    ):
        # Creating temporary file
        self.file = QTemporaryFile()
        self.file.open()
        self.dictionary = DictWithPWL(
            "en_US",
            self.file.fileName(),
        )

        self.addToDictionary = addToDictionary

        self.word_list = set(personal_word_list)
        self.load_words()

    def load_words(self):
        for word in self.word_list:
            self.dictionary.add(word)

    def suggestions(self, word: str) -> list[str]:
        return self.dictionary.suggest(word)

    def correction(self, word: str) -> str:
        return self.dictionary.suggest(word)[0]

    def add(self, new_word: str) -> bool:
        if self.check(new_word):
            return False
        self.word_list.add(new_word)
        self.addToDictionary(new_word)
        self.dictionary.add(new_word)
        return True

    def check(self, word: str) -> bool:
        return self.dictionary.check(word)

    def getNewWords(self) -> set[str]:
        return self.word_list



def find_word_bounds(text: str, cursor_pos: int) -> tuple[int, int]:
    """Find the start and end of the word under the cursor."""
    start = cursor_pos
    while start > 0 and text[start - 1].isalnum():
        start -= 1

    end = cursor_pos
    while end < len(text) and text[end].isalnum():
        end += 1

    return start, end
